#include <sys/time.h>
#if 0
#include <sys/resource.h>
#include <sys/wait.h>
#endif
#include <sys/errno.h>
#include <string.h>
#ifdef RS6000
#include <rpc/types.h>
#endif
#include <rpc/xdr.h>

#include "inet.h"

#define QLEN    5
#define BUFSIZE    4096

/* Funcitons in this file */
int Server(char *, char *, char *, char *);
int BindServer(char *, char *, char *, char *);
int AcceptClient(int );
int AcceptControl(int);
int Client(char *, char *, char *, char *);
int ClientInit(char *, char *);
int SendDisplay(unsigned char *, int , int );
int ReadBytes(int , char *, int );
int WriteBytes(int , char *, int );
int ReadInt(int, int *, int);
int ReadDouble(int, double *, int);
int WriteInt(int, int *, int);
int WriteDouble(int, double *, int);

/* Globals - there's got to be a way around this! Think! */
static int port;
static int psock;               /* socket descriptor */
static struct sockaddr_in  serv_addr;




/*--------------------------------------------------------------------------
Name: int Server(char *hostname, char *service, char *protocol, char *channel)
In: None
Out: sockfd
Comments: This function binds the server data/control sockets and accepts
          connection to socket. Returns -1 in case connect fails.
--------------------------------------------------------------------------*/
int
Server(char *hostname, char *service, char *protocol, char *channel) {

  int  csock;           /* child socket */
  int  clilen;          /* client socket */
  struct sockaddr_in   serv_addr, cli_addr;

  if (!strcmp(channel, "data"))
    port = SERV_DATA_PORT;
  else if (!strcmp(channel, "control"))
    port = SERV_CONTROL_PORT;
  else if (!strcmp(channel, "map_segment"))
    port = MAPPER_DATA_PORT;


  psock=OpenSock(hostname, service, protocol, &serv_addr, port);
  printf("Binding server on port: %hu at IP address: %lu \n",
         serv_addr.sin_port, serv_addr.sin_addr); 
  if (bind(psock, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
    printf("ERROR: Can't bind server\n");
    return(ERROR);
  }

  if (listen(psock, QLEN) < 0) {
    printf("ERROR: Can't listen\n");
    return(ERROR);
  }

  clilen = sizeof(struct sockaddr_in) ;
  csock=accept(psock, (struct sockaddr *)&cli_addr, &clilen);
  if (csock < 0) {
    printf("ERROR: Accept error\n");
    return(ERROR);
  }
  close(psock);
  return(csock);

}




/*--------------------------------------------------------------------------
Name: int BindServer(char *hostname, char *service, char *protocol, 
                     char *channel)
In: None
Out: sockfd
Comments: This function binds the server data/control sockets.
--------------------------------------------------------------------------*/
int BindServer(char *hostname, char *service, char *protocol, 
	       char *channel) {

  int  csock;           /* child socket */
  struct sockaddr_in   serv_addr;

  if (!strcmp(channel, "data"))
    port = SERV_DATA_PORT;
  else
    port = SERV_CONTROL_PORT;

  psock=OpenSock(hostname, service, protocol, &serv_addr, port);
  printf("Binding server on port: %hu at IP address: %lu \n",
         serv_addr.sin_port, serv_addr.sin_addr); 
  if (bind(psock, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
    printf("ERROR: Can't bind server\n");
    return(ERROR);
  }

  if (listen(psock, QLEN) < 0) {
    printf("ERROR: Can't listen\n");
    return(ERROR);
  }

  return(psock);

}




/*--------------------------------------------------------------------------
Name: int AcceptClient(int psock)
In: None
Out: sockfd
Comments: This function accepts connection from client. Using concurrent
          server model.
--------------------------------------------------------------------------*/
int AcceptClient(int psock) {

  int                  csock;
  int                  clilen;
  struct sockaddr_in   cli_addr;
  int                  childpid;

  clilen = sizeof(struct sockaddr_in) ;
  csock=accept(psock, (struct sockaddr *)&cli_addr, &clilen);
  if (csock < 0) {
    printf("ERROR: Data channel accept error\n");
    return(ERROR);
  }

  if ((childpid = fork()) < 0) {
    printf("ERROR: fork error\n");
    return(ERROR);
  } else if (childpid == 0) { /* child process */
    return(csock);
  } else {                    /* parent process */
    close(csock);
    return(NULL);
  }

  return(OK);

}



/*--------------------------------------------------------------------------
Name: int AcceptControl(int psock)
In: None
Out: sockfd
Comments: This function accepts control connection from client.
--------------------------------------------------------------------------*/
int AcceptControl(int psock) {

  int csock;
  int clilen;
  struct sockaddr_in   cli_addr;

  clilen = sizeof(struct sockaddr_in) ;
  csock=accept(psock, (struct sockaddr *)&cli_addr, &clilen);
  if (csock < 0) {
    printf("ERROR: Control channel accept error\n");
    return(ERROR);
  }

  close (psock);
  return(csock);

}




/*--------------------------------------------------------------------------
Name: int Client(char *hostname, char *service, char *protocol
                 char *channel)
In: None
Out: sockfd
Comments: This function connects to the server data/control sockets. 
          Retries connection at every 1 sec. interval, timeout is (soft)set
          to 10 attempts.
          Returns -1 in case connect fails.
--------------------------------------------------------------------------*/
int Client(char *hostname, char *service, char *protocol, char *channel) { 

  struct sockaddr_in  serv_addr;
  int oldmask;
  int count;

  if (!strcmp(channel, "data"))
    port = SERV_DATA_PORT;
  else if (!strcmp(channel, "control"))
    port = SERV_CONTROL_PORT;
  else if (!strcmp(channel, "map_segment"))
    port = MAPPER_DATA_PORT;

  count = 0;
  do {
     psock=OpenSock(hostname, service, protocol, &serv_addr, port);
     if (connect(psock, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) == 0)
     {
        break;
     }
     else
     {
        count++;
        if (count < TIMEOUT_N)
        {
           printf("Retrying connection \n");
/*         sleep(1); */
#define SLEEPTIME 100
          {int i;
             for (i=0;i<SLEEPTIME;i++) {
                double sleepvar = 681.231;
                sleepvar = 1.0 / sleepvar;
             }
          }
#undef SLEEPTIME
        }
     }
  } while (count < TIMEOUT_N);
  
  if (count == TIMEOUT_N) {
     printf("ERROR: Failed to connect to server: %s at port: %hu \
           after %d attempts \n", hostname, serv_addr.sin_port, count);
     return(ERROR);
  }

  return (psock);
}




/*--------------------------------------------------------------------------
Name: int ClientInit(char *hostname, char *channel)
In: None
Out: sockfd
Comments: This function connects to the server data/control sockets.
          Returns -1 in case connect fails.
--------------------------------------------------------------------------*/
int ClientInit(char *hostname, char *channel) {

   int sockfd;

   printf("Connecting %s channel\n", channel);
   if ((sockfd=Client(hostname, "tcp", "tcp", channel)) == ERROR) {
     printf("ERROR in connecting %s channel\n", channel);
     return(ERROR);
   }
   printf("%s channel established\n", channel);

   return (sockfd);

}




/*--------------------------------------------------------------------------
NOTE: Implementing blocking read. Require that message length from client match the size of view defined in view.h
--------------------------------------------------------------------------*/

/*
int
GetView(sockfd, view)

int sockfd;
char *view;

{
int nread, status = 1;
char *flag;

flag = (char *)malloc(100);
if((nread = readn(sockfd, view, VIEWLEN)) != VIEWLEN) {
  printf("ERROR server: readn error\n");
  exit (ERROR);
}
sscanf(view, "%s", flag);
if (strcmp(flag, "EOT") == 0)
  status = NULL;

return status;
}
*/





/*-----------------------------------------------------------------------
Function to send image to the front-end 
------------------------------------------------------------------------*/

int
SendDisplay(unsigned char *buf, int size, int sockfd)

{
  return (WriteBytes(sockfd, (char *) buf, size));
}




/*-----------------------------------------------------------------------
Name: ReadBytes(int sockfd, char *message, int size)
In: signature
Out: message from socket
Comments: Function to read message from a socket using XDR.
------------------------------------------------------------------------*/
int ReadBytes(int sockfd, char *message, int size) {

  XDR xdrs;
  char *buf;
  int nbytes;

  buf = (char *)malloc(size+16);
  xdrmem_create(&xdrs, buf, size+16, XDR_DECODE);
  if ((nbytes = readn(sockfd, buf, size)) != size)
    return (ERROR);
  else {
    if (xdr_bytes(&xdrs, &message, &size, size) == FALSE) {
      printf("ERROR in xdr conversion\n");
      xdr_destroy(&xdrs);
      return (ERROR);
    }
  }
 
  xdr_destroy(&xdrs);
  free(buf);
  return (nbytes);
}




/*-----------------------------------------------------------------------
Name: WriteBytes(int sockfd, char *message, int size)
In: signature
Out: message to socket
Comments: Function to write message to a socket using XDR.
------------------------------------------------------------------------*/
int WriteBytes(int sockfd, char *message, int size) {

  int nbytes = 0;
  XDR xdrs;
  char *buf;

  buf = (char *)malloc(size+16);
  xdrmem_create(&xdrs, buf, size+16, XDR_ENCODE);
  if (xdr_bytes(&xdrs, &message, &size, size) == FALSE) {
    printf("ERROR in xdr conversion \n");
    xdr_destroy(&xdrs);
    return (ERROR);
  }
  if ((nbytes = writen(sockfd, buf, size)) != size) {
    xdr_destroy(&xdrs);
    return(ERROR);
  }

  xdr_destroy(&xdrs);
  free(buf);
  return(nbytes);

}





/*-----------------------------------------------------------------------
Name: ReadBytes(int sockfd, char *message, int size)
In: signature
Out: message from socket
Comments: Function to read message from a socket using XDR.
------------------------------------------------------------------------*/
int ReadInt(int sockfd, int *message, int size)
{
   XDR             xdrs;
   int             buf_size;
   char           *buf;
   int             nbytes;
   int             Condition;
   
   buf_size = size * BYTES_PER_XDR_UNIT;
   if ((buf = (char *) malloc(buf_size)) == NULL)
   {
       printf("ReadInt2Float(): Out of memory!\n");
       exit(1);
   }

   xdrmem_create(&xdrs, buf, buf_size, XDR_DECODE);

   if ((nbytes = readn(sockfd, buf, buf_size)) == buf_size) {
     if (xdr_vector(&xdrs, message, size, sizeof(int), 
		    xdr_int) == TRUE) {
	   Condition = size;
	 }
	 else {
       printf("ERROR in xdr conversion\n");
       Condition = ERROR;
     }
   }
   else
      Condition = ERROR;
      
   xdr_destroy(&xdrs);
   free(buf);
   return(Condition);
 }




/*-----------------------------------------------------------------------
Name: ReadBytes(int sockfd, char *message, int size)
In: signature
Out: message from socket
Comments: Function to read message from a socket using XDR.
------------------------------------------------------------------------*/
int ReadInt2Float(int sockfd, float *message, int size) {

   XDR             xdrs;
   int             buf_size;
   char            *buf;
   int             nbytes, i, elem_processed = 0, err_flag = FALSE;
   short           temp;

   buf_size = size * BYTES_PER_XDR_UNIT;
   if ((buf = (char *) malloc(buf_size)) == NULL)
   {
       printf("ReadInt2Float(): Out of memory!\n");
       exit(1);
   }
   
   xdrmem_create(&xdrs, buf, buf_size, XDR_DECODE);

   if ((nbytes = readn(sockfd, buf, buf_size)) == buf_size) {
       for (i = 0; i < size; i++)
       {
          if (xdr_short(&xdrs, &temp) == TRUE) {
             message[i] = (float)temp;
          }
          else {
	         printf("ReadInt2Float(): ERROR in xdr conversion\n");
	         i = ERROR;
	         break;
	      }
	   }   
   }
   else
      i = ERROR;

   xdr_destroy(&xdrs);
   free(buf);
   
   return(i);       

}




/*-----------------------------------------------------------------------
Name: ReadBytes(int sockfd, char *message, int size)
In: signature
Out: message from socket
Comments: Function to read message from a socket using XDR.
------------------------------------------------------------------------ */
int ReadDouble(int sockfd, double *message, int size)
{
   XDR             xdrs;
   int             buf_size;
   char           *buf;
   int             nbytes;
   int             Condition;

   buf_size = size * sizeof(double);
   if ((buf = (char *) malloc(buf_size)) == NULL)
   {
       printf("ReadInt2Float(): Out of memory!\n");
       exit(1);
   }
   
   xdrmem_create(&xdrs, buf, buf_size, XDR_DECODE);
   if ((nbytes = readn(sockfd, buf, buf_size)) == buf_size) {
      if (xdr_vector(&xdrs, message, size, sizeof(double), 
		    xdr_double) == TRUE) {
	     Condition = size;
	  }
	  else {
         printf("ERROR in xdr conversion\n");
         Condition = ERROR;
      }
   }
   else
      Condition = ERROR;
      
   xdr_destroy(&xdrs);
   free(buf);
   return(nbytes / sizeof(double));

}




/*-----------------------------------------------------------------------
Name: WriteBytes(int sockfd, char *message, int size)
In: signature
Out: message to socket
Comments: Function to write message to a socket using XDR.
------------------------------------------------------------------------*/
int WriteInt(int sockfd, int *message, int size) {

  int nbytes = 0;
  XDR xdrs;
  int             buf_size;
  char           *buf;


   buf_size = size * BYTES_PER_XDR_UNIT;
   if ((buf = (char *) malloc(buf_size)) == NULL)
   {
       printf("WriteInt(): Out of memory!\n");
       exit(1);
   }
  xdrmem_create(&xdrs, buf, buf_size, XDR_ENCODE);
  if (xdr_vector(&xdrs, message, size, sizeof(int), xdr_int) == FALSE) {
    printf("ERROR in xdr conversion \n");
    xdr_destroy(&xdrs);
    return (ERROR);
  }

  if ((nbytes = writen(sockfd, buf, buf_size)) != buf_size) {
    xdr_destroy(&xdrs);
    return(ERROR);
  }

  xdr_destroy(&xdrs);
  free(buf);
  return(size);

}





/*-----------------------------------------------------------------------
Name: WriteDouble(int sockfd, char *message, int size)
In: signature
Out: message to socket
Comments: Function to write message to a socket using XDR.
------------------------------------------------------------------------*/
int WriteDouble(int sockfd, double *message, int size) {

  int nbytes = 0;
  XDR xdrs;
  int             buf_size;
  char            *buf;

   buf_size = size * sizeof(double);
   if ((buf = (char *) malloc(buf_size)) == NULL)
   {
       printf("ReadInt2Float(): Out of memory!\n");
       exit(1);
   }

  xdrmem_create(&xdrs, buf, buf_size, XDR_ENCODE);
  if (xdr_vector(&xdrs, message, size, sizeof(double), xdr_double) == FALSE) {
    printf("ERROR in xdr conversion \n");
    xdr_destroy(&xdrs);
    return (ERROR);
  }

  if ((nbytes = writen(sockfd, buf, buf_size)) != buf_size) {
    xdr_destroy(&xdrs);
    return(ERROR);
  }

  xdr_destroy(&xdrs);
  free(buf);
  return(size);

}
